#include <iostream>
#include <iomanip>
#include "co2co2.hpp"
#include "unit_system.hpp"
#include "helper_functions.hpp"

using namespace h2ococo2forcefields;
using namespace h2ococo2forcefields::unit_system;


Co2Co2::Co2Co2() : PotentialBase(),
                   qcqc(parameters::co2_qc_*parameters::co2_qc_),
                   qcqo(parameters::co2_qc_*parameters::co2_qo_),
                   qoqo(parameters::co2_qo_*parameters::co2_qo_)
{ }


char const * Co2Co2::getName() {
    return "Co2Co2";
}

void Co2Co2::computeCO2Intra(const int nCo2, Co2 *const *const co2List, double & energy, const double* const diagbox) {
    Co2 *co2;
    for(int i=nCo2-1;i>=0;--i) {
        co2 = co2List[i];
        calculateInternalGeometry(*co2);
        if(co2->fixed_) continue;
        intraMolecular(*co2,energy);
    }
    return;
}

void Co2Co2::computeCO2_CO2(const int nCo2, Co2 *const *const co2List, double &energy, const double* const diagbox) {

    setPeriodicity(diagbox);
    double r[3],r1,r2;

    Co2 *co2_1, *co2_2;

    double e_intra = energy;


    //set internal geometry and calculate intramolecular interactions.
    for(int i=nCo2-1; i>=0; --i)
    {
        //calculate internal geometry
        co2_1 = co2List[i];
        calculateInternalGeometry(*co2_1);

        if(co2_1->fixed_) continue;
    
        intraMolecular(*co2_1,energy);
    }
    e_intra = energy - e_intra; 
    
    double e_inter = energy;    
    //intermolecular interactions
    for(int i=nCo2-1; i >= 0; --i) {            
        co2_1 = co2List[i];
        for(int j=i-1; j>=0; --j) {        
            co2_2 = co2List[j];
            if(co2_1->fixed_ & co2_2->fixed_) continue;                      

            distance(co2_1->rx_,co2_2->rx_,r,r1,r2);
            
            if(r1 <= cutoff_ - switchingWidth_) {
                interMolecular(*co2_1,*co2_2,energy);
            }   
            else if(r1 < cutoff_) {
                //we need to create to new co molecules. 
                double fSwitch_1[4][3]={0.0}, fSwitch_2[4][3]={0.0};
                double U=0.0;

                Co2 co2switch_1(*co2_1,fSwitch_1[0],fSwitch_1[1],fSwitch_1[2],fSwitch_1[3]);
                Co2 co2switch_2(*co2_2,fSwitch_2[0],fSwitch_2[1],fSwitch_2[2],fSwitch_2[3]);

                interMolecular(co2switch_1, co2switch_2, U);

                switching(co2switch_1,co2switch_2,U,r,r1);
      
                co2_1->incrementForces(co2switch_1);
                co2_2->incrementForces(co2switch_2);
                energy+=U;
            }
        }
    }
    e_inter = energy - e_inter;
    //std::cout.precision(6);
    //std::cout<<std::fixed<<"CO2-CO2\n\tTotal: "<<e_inter+e_intra<<"\n\tInter: "<<e_inter<<"\n\tIntra: "<<e_intra<<std::endl;
}

void Co2Co2::computeInterMolecularCo2Co2(Co2 &co21, Co2 &co22, double &energy,double *force,const double* const diagbox) {
    //calculate internal geometry
    
    co21.resetForces();
    co22.resetForces();
    energy =0.0;
    double r[3],r1,r2;
    
    distance(co21.rx_,co22.rx_,r,r1,r2);
    if(r1 >= cutoff_) return;
    
    interMolecular(co21,co22,energy);
    if(r1 > ( cutoff_ - switchingWidth_ ) ) switching(co21, co22, energy,r,r1);
    
    for(int i=2;i>=0;--i) force[i]+=co21.fc_[i]+co21.fo1_[i]+co21.fo2_[i]+co21.fx_[i];
    return;
}


inline void Co2Co2::interMolecular(Co2 &co2_1,Co2 &co2_2, double &energy) {

    coulomb(co2_1.rc_, co2_2.rc_, co2_1.fc_, co2_2.fc_, energy, qcqc);
    coulomb(co2_1.rc_, co2_2.ro1_, co2_1.fc_, co2_2.fo1_, energy, qcqo);
    coulomb(co2_1.rc_, co2_2.ro2_, co2_1.fc_, co2_2.fo2_, energy, qcqo);
    coulomb(co2_1.ro1_, co2_2.rc_, co2_1.fo1_, co2_2.fc_, energy, qcqo);
    coulomb(co2_1.ro2_, co2_2.rc_, co2_1.fo2_, co2_2.fc_, energy, qcqo);
    coulomb(co2_1.ro1_, co2_2.ro1_, co2_1.fo1_, co2_2.fo1_, energy, qoqo);        
    coulomb(co2_1.ro1_, co2_2.ro2_, co2_1.fo1_, co2_2.fo2_, energy, qoqo); 
    coulomb(co2_1.ro2_, co2_2.ro1_, co2_1.fo2_, co2_2.fo1_, energy, qoqo);               
    coulomb(co2_1.ro2_, co2_2.ro2_, co2_1.fo2_, co2_2.fo2_, energy, qoqo); 

    lennardJones(co2_1.rc_,  co2_2.rc_,  co2_1.fc_,  co2_2.fc_, energy, parameters::co2co2_epsiloncc_, parameters::co2co2_sigmacc_);
    lennardJones(co2_1.rc_,  co2_2.ro1_, co2_1.fc_,  co2_2.fo1_,energy, parameters::co2co2_epsilonco_, parameters::co2co2_sigmaco_);
    lennardJones(co2_1.rc_,  co2_2.ro2_, co2_1.fc_,  co2_2.fo2_,energy, parameters::co2co2_epsilonco_, parameters::co2co2_sigmaco_);
    lennardJones(co2_1.ro1_, co2_2.rc_,  co2_1.fo1_, co2_2.fc_, energy, parameters::co2co2_epsilonco_, parameters::co2co2_sigmaco_);
    lennardJones(co2_1.ro2_, co2_2.rc_,  co2_1.fo2_, co2_2.fc_, energy, parameters::co2co2_epsilonco_, parameters::co2co2_sigmaco_);
    lennardJones(co2_1.ro1_, co2_2.ro1_, co2_1.fo1_, co2_2.fo1_,energy, parameters::co2co2_epsilonoo_, parameters::co2co2_sigmaoo_);
    lennardJones(co2_1.ro1_, co2_2.ro2_, co2_1.fo1_, co2_2.fo2_,energy, parameters::co2co2_epsilonoo_, parameters::co2co2_sigmaoo_);
    lennardJones(co2_1.ro2_, co2_2.ro1_, co2_1.fo2_, co2_2.fo1_,energy, parameters::co2co2_epsilonoo_, parameters::co2co2_sigmaoo_);
    lennardJones(co2_1.ro2_, co2_2.ro2_, co2_1.fo2_, co2_2.fo2_,energy, parameters::co2co2_epsilonoo_, parameters::co2co2_sigmaoo_);
   
}




/************************************************
Calculate intramolecular forces
*************************************************/

void Co2Co2::intraMolecular(Co2 &co2, double &energy) {
    double dtheta,theta,c11,c12,c22,P,f1,f2,dr1,dr2;   
    double crossProd[3], crossProd_1; 
    
    dr1 = co2.d1_1_ - parameters::co2_re_;
    dr2 = co2.d2_1_ - parameters::co2_re_;

    c11 = co2.d1_1_*co2.d1_1_;
    c22 = co2.d2_1_*co2.d2_1_;
    c12 = helper_functions::dotProduct(co2.d1_,co2.d2_);

    helper_functions::crossProduct(co2.d1_,co2.d2_,crossProd);
    crossProd_1 = helper_functions::norm(crossProd);
    dtheta = atan2( crossProd_1 , c12 ) - pi ;    
    energy += parameters::co2_k1_*0.5*(dr1*dr1+dr2*dr2) + parameters::co2_k2_*dr1*dr2 + parameters::co2_ktheta_*0.5*dtheta*dtheta;

    if(fabs(dtheta) < 1e-8) {
        dtheta*=dtheta; //dtheta^2
        P = parameters::co2_ktheta_*( -1.0 - 1.0/6 * dtheta * ( 1.0 + 7.0/60 * dtheta * ( 1.0 + 31.0/294 * dtheta ) ) ) / ( co2.d1_1_ * co2.d2_1_ );  
    }
    else {
        P = parameters::co2_ktheta_*dtheta/crossProd_1; 
    }

    for(int i=2;i>=0;--i) {
        f1 = P * ( co2.d2_[i] - co2.d1_[i]*c12/c11 ) - parameters::co2_k2_*dr2*co2.d1_[i]/co2.d1_1_ - parameters::co2_k1_*dr1*co2.d1_[i]/co2.d1_1_;
        f2 = P * ( co2.d1_[i] - co2.d2_[i]*c12/c22 ) - parameters::co2_k2_*dr1*co2.d2_[i]/co2.d2_1_ - parameters::co2_k1_*dr2*co2.d2_[i]/co2.d2_1_;

        co2.fo1_[i] += f1;
        co2.fo2_[i] += f2;
        co2.fc_[i] -= f1 + f2;
    }
}

void Co2Co2::calculateInternalGeometry(Co2 &co2) {
    double d1[3],d2[3],d1_1,d2_1,d1_2,d2_2;
    distance(co2.ro1_,co2.rc_,d1,d1_1,d1_2);
    distance(co2.ro2_,co2.rc_,d2,d2_1,d2_2);
    co2.setD(d1,d2,d1_1,d2_1);
}


