#include <iostream>
#include "coco.hpp"
#include "unit_system.hpp"


using namespace h2ococo2forcefields;
using namespace h2ococo2forcefields::unit_system;

CoCo::CoCo() : PotentialBase() 
{}


char const * CoCo::getName()
{
      return "CoCo";
}

void CoCo::computeCOIntra(const int nCo, Co *const *const coList, double & energy, const double* const diagbox){
    Co *co;
    for(int i=nCo-1;i>=0;--i) {
        co = coList[i];
        if(co->fixed_) continue;
        morse(co->rc_, co->ro_, co->fc_, co->fo_, co->d_, co->d_1_, energy, parameters::co_beta_, parameters::co_D_, parameters::co_re_);
    }
    return;
}

void CoCo::computeC_O_(const int nCo,Co *const *const coList, double & energy, const double * const diagbox) {
    setPeriodicity(diagbox);
    
    double r[3],r1,r2;
    Co *co1, *co2;

    double e_intra = energy;

    for(int i=nCo-1; i>=0; --i) {
        co1 = coList[i];
        calculateInternalGeometry(*co1);
    
        if(co1->fixed_) continue;
        
        morse(co1->rc_, co1->ro_, co1->fc_, co1->fo_, co1->d_, co1->d_1_, energy, parameters::co_beta_, parameters::co_D_, parameters::co_re_);
    }
    e_intra = energy - e_intra;
    double e_inter = energy;

    for (int i=nCo-1; i >= 0; --i) {
        co1 = coList[i];
	
        for (int j=i-1; j >= 0; --j) {
            co2 = coList[j]; 
            if(co1->fixed_ & co2->fixed_) continue;        
        
            distance(co1->rx_,co2->rx_,r,r1,r2);
            if(r1 <= cutoff_ - switchingWidth_) {
                //Full interactions
                coulombFull(*co1,*co2,energy);
                buckinghamFull(*co1,*co2,energy);
            }

            else if(r1 < cutoff_) {
                //switched interactions
                double f1[3][3]={0.0}, f2[3][3]={0.0}; ///<temporary CO forces
                double U=0.0;	///< temporary energy
                      		
                Co co_tmp1(*co1,f1[0],f1[1],f1[2]);
          		Co co_tmp2(*co2,f2[0],f2[1],f2[2]);
                //Calculate coulomb and LJ interactions.                        
                coulombFull(co_tmp1,co_tmp2,U);
                buckinghamFull(co_tmp1, co_tmp2, U);   

                //perform switching
                switching(co_tmp1,co_tmp2,U,r,r1);
                co1->incrementForces(co_tmp1);
                co2->incrementForces(co_tmp2);
                energy+=U;	
            }
        }
    }
/*
    e_inter = energy - e_inter;
    std::cout.precision(6);
    std::cout<<std::fixed<<"CO-CO\n\tTotal: "<<e_inter+e_intra<<"\n\tInter: "<<e_inter<<"\n\tIntra: "<<e_intra<<std::endl;
*/
}


void CoCo::computeInterMolecularCoCo(Co &co1, Co &co2, double &energy,double *force,const double* const diagbox) {
    double r[3],r1,r2;
    
    co1.resetForces();
    co2.resetForces();
    energy =0.0;
    distance(co1.rx_,co2.rx_,r,r1,r2);
    if(r1 >= cutoff_) return;
    
    coulombFull(co1,co2,energy);
    buckinghamFull(co1,co2,energy);
    if(r1 >= (cutoff_-switchingWidth_)) switching(co1,co2,energy,r,r1);
    
    for(int i=2;i>=0;--i) force[i] += co1.fc_[i]+co1.fo_[i]+co1.fx_[i];
}


//calculate buckingham intermolecules potential between two co molecules. 
void CoCo::buckinghamFull(Co & co1, Co & co2, double & energy) {
    //O-O interaction
    buckingham(co1.ro_,co2.ro_,co1.fo_,co2.fo_,energy, parameters::coco_csixoo_, parameters::coco_cbuckoo_, parameters::coco_cexpoo_);
    //C-C interaction
    buckingham(co1.rc_,co2.rc_,co1.fc_,co2.fc_,energy, parameters::coco_csixcc_, parameters::coco_cbuckcc_, parameters::coco_cexpcc_);
    //C-O interactions
    buckingham(co1.rc_,co2.ro_,co1.fc_,co2.fo_,energy, parameters::coco_csixco_, parameters::coco_cbuckco_, parameters::coco_cexpco_);
    buckingham(co1.ro_,co2.rc_,co1.fo_,co2.fc_,energy, parameters::coco_csixco_, parameters::coco_cbuckco_, parameters::coco_cexpco_);
}

void CoCo::coulombFull(Co & co1, Co & co2, double & U)
{

	double dqo1dr[3],dqo2dr[3],dqc1dr[3],dqc2dr[3],dqx1dr[3],dqx2dr[3];
	double qo1,qo2,qc1,qc2,qx1,qx2;
	double r[3],r1,r2;
	double en;

	qc1=parameters::co_qc_*exp( -parameters::co_sigmac_*(co1.d_1_-parameters::co_re_) );
	qo1=parameters::co_qo_*exp( -parameters::co_sigmao_*(co1.d_1_-parameters::co_re_) );
	qc2=parameters::co_qc_*exp( -parameters::co_sigmac_*(co2.d_1_-parameters::co_re_) );
	qo2=parameters::co_qo_*exp( -parameters::co_sigmao_*(co2.d_1_-parameters::co_re_) );
	qx1 = -qc1-qo1;
	qx2 = -qc2-qo2;
	
	//calculate 'charge derivatives'
	for(int i=2;i>=0;--i) {
		dqo1dr[i]=parameters::co_sigmao_*co1.d_[i]/co1.d_1_;
		dqo2dr[i]=parameters::co_sigmao_*co2.d_[i]/co2.d_1_;
		
		dqc1dr[i]=parameters::co_sigmac_*co1.d_[i]/co1.d_1_;
		dqc2dr[i]=parameters::co_sigmac_*co2.d_[i]/co2.d_1_;
		
		dqx1dr[i]=ONE_OVER_4_PI_EPSILON0*(parameters::co_sigmac_*qc1+parameters::co_sigmao_*qo1)*co1.d_[i]/co1.d_1_;
		dqx2dr[i]=ONE_OVER_4_PI_EPSILON0*(parameters::co_sigmac_*qc2+parameters::co_sigmao_*qo2)*co2.d_[i]/co2.d_1_;
	}

	
	//Coulomb interactions between carbons. 
		distance(co1.rc_, co2.rc_, r,r1,r2); //calculate distance. r=co1.rc_-ro2.rc_;
		en=ONE_OVER_4_PI_EPSILON0*qc1*qc2/r1;
		U+=en;	
		//calculate forces
		for(int i=0; i<3;i++) {
			co1.fc_[i]+=en*r[i]/r2 + en*dqc1dr[i] ;
			co1.fo_[i]+=-en*dqc1dr[i]; 
			
			co2.fc_[i]+=-en*r[i]/r2 + en*dqc2dr[i];
			co2.fo_[i]+=-en*dqc2dr[i];
		}

	//Coulomb interactions between oxygens.
		distance(co1.ro_,co2.ro_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qo1*qo2/r1;
		U+=en;
		//calculate forces
		for(int i=0; i<3; i++) {
			co1.fc_[i]+=en*dqo1dr[i];
			co1.fo_[i]+=en*r[i]/r2 - en*dqo1dr[i];
			
			co2.fc_[i]+=en*dqo2dr[i];
			co2.fo_[i]+=-en*r[i]/r2 - en*dqo2dr[i];
		
		}
	
	//Coulomb interaction between centers of mass
		distance(co1.rx_,co2.rx_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qx1*qx2/r1;
		U+=en;
		//calculate forces
		for(int i=0; i<3;i++) {
			co1.fc_[i]+=-qx2*dqx1dr[i]/r1;
			co1.fo_[i]+= qx2*dqx1dr[i]/r1;
			co1.fx_[i]+=en*r[i]/r2;
			
			co2.fc_[i]+=-qx1*dqx2dr[i]/r1;
			co2.fo_[i]+= qx1*dqx2dr[i]/r1;
			co2.fx_[i]+=-en*r[i]/r2;
		}
		
	//Coulomb intractions between C and X
		//pair 1
		distance(co1.rc_,co2.rx_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qc1*qx2/r1;
		U+=en;
		//calculate forces
		for(int i=0; i<3; i++) {
			co1.fc_[i]+=en*r[i]/r2 + en*dqc1dr[i];
			co1.fo_[i]+=-en*dqc1dr[i];
			
			co2.fc_[i]+=-qc1*dqx2dr[i]/r1;
			co2.fo_[i]+= qc1*dqx2dr[i]/r1;
			co2.fx_[i]+= -en*r[i]/r2;
		}	
	
		//pair 2
		distance(co1.rx_,co2.rc_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qx1*qc2/r1;
		U+=en;
		//calculate forces
		for(int i=0;i<3;i++) {
			co1.fc_[i]+=-dqx1dr[i]*qc2/r1;
			co1.fo_[i]+= dqx1dr[i]*qc2/r1;
			co1.fx_[i]+=en*r[i]/r2;
			
			co2.fc_[i]+=-en*r[i]/r2+en*dqc2dr[i];
			co2.fo_[i]+= -en*dqc2dr[i];
		}
		
	//Coulomb interactions between O and X 
		//pair 1
		distance(co1.ro_,co2.rx_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qo1*qx2/r1;
		U+=en;
		//calculate forces
		for(int i=0;i<3;i++) {
			co1.fc_[i]+=en*dqo1dr[i];
			co1.fo_[i]+=en*r[i]/r2-en*dqo1dr[i];
			
			co2.fc_[i]+=-dqx2dr[i]*qo1/r1;
			co2.fo_[i]+= dqx2dr[i]*qo1/r1;
			co2.fx_[i]+=-en*r[i]/r2;
		}
		
		//pair 2
		distance(co1.rx_,co2.ro_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qx1*qo2/r1;
		U+=en;
		//calculate forces
		for(int i=0;i<3;i++) {
			co1.fc_[i]+=-dqx1dr[i]*qo2/r1;
			co1.fo_[i]+= dqx1dr[i]*qo2/r1; 
			co1.fx_[i]+= en*r[i]/r2;
			
			co2.fc_[i]+= en*dqo2dr[i];
			co2.fo_[i]+= -en*r[i]/r2-en*dqo2dr[i];
		}
		
	//Coulomb interaction between O and C
		//pair 1
		distance(co1.rc_,co2.ro_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qc1*qo2/r1;
		U+=en;
		//calculate forces
		for(int i=0;i<3;i++) {
			co1.fc_[i]+= en*r[i]/r2 + en*dqc1dr[i];
			co1.fo_[i]+= -en*dqc1dr[i];
			
			co2.fc_[i]+= en*dqo2dr[i];
			co2.fo_[i]+= -en*r[i]/r2 - en*dqo2dr[i];
		}
		
		//pair 2
		distance(co1.ro_,co2.rc_,r,r1,r2);
		en=ONE_OVER_4_PI_EPSILON0*qo1*qc2/r1;
		U+=en;
		//calculate forces
		for(int i=0;i<3;i++) {
			co1.fc_[i]+= en*dqo1dr[i];
			co1.fo_[i]+= en*r[i]/r2 - en*dqo1dr[i];
			
			co2.fc_[i]+= -en*r[i]/r2 + en*dqc2dr[i];
			co2.fo_[i]+= -en*dqc2dr[i];
		}
}

void CoCo::calculateInternalGeometry(Co & co) {
    double r[3], r1;
    distance(co.rc_, co.ro_, r, r1);
    co.setD(r,r1);
}


void CoCo::calcCO_charge(Co & co, double & qc, double & qo) {
	double r1;        
	distance(co.rc_,co.ro_,r1);
	qc=parameters::co_qc_*exp( -parameters::co_sigmac_*(r1-parameters::co_re_) );
	qo=parameters::co_qo_*exp( -parameters::co_sigmao_*(r1-parameters::co_re_) );

}

