#include "geometry.hpp"
#include <iostream>
#include <cmath>
#include <cassert>
using namespace h2ococo2forcefields;


Geometry::Geometry() {

}


Geometry::Geometry(const double* const p) {
    periods_[0] = p[0];
    periods_[1] = p[1];
    periods_[2] = p[2];
    inv_periods_[0] = 1.0/p[0];
    inv_periods_[1] = 1.0/p[1];
    inv_periods_[2] = 1.0/p[2];
}

/**
@return the lattice translation vector between the boxes in which r2 is located, with respect to r1.
*/
void Geometry::returnCellVector(const double *r1,const double *r2, double *t) 
{
    double r[3],x;
    //first calculate the distance vector from r=r1-r2, from  r2 to r1, using the pbc.
    distance(r1,r2,r);
    for(int i=2;i>=0;--i) {
        x = (r1[i]-r[i])*inv_periods_[i];
        if(x < 0.0) t[i] = -periods_[i];
        else if(x > 1.) t[i] = periods_[i];
        else t[i]=0.0;
    }
    
}

/** Minimum image representation.
@return Minimum image representation of @a r. The value returned is @f$ \frac{-period}{2} \le r \le \frac{period}{2} @f$
@param[in] r Cyclic coordinate.
@param[in] period Period of the coordinates.
*/
double Geometry::applyPeriodicity0(double r, double const period)
{
    if (not isinf(period)) {
        double n=r/period+0.5;
        // This is slightly more efficient than using function floor().
        int m=(int)n;
        if (n < 0)
            -- m;
        r-=m*period;
    };
    return r;
}

/** Minimum image representation.
@return Minimum image representation of @a r. The value returned is @f$ \frac{-period}{2} \le r \le \frac{period}{2} @f$.
@param[in] r      Three-dimension vector.
@param[in] periods      Three-dimension vector containing periods along each of the axes.
*/
void Geometry::applyPeriodicity0(double r[], double const periods[])
{
    for (int i=2; i >= 0; --i)
        r[i]=applyPeriodicity0(r[i], periods[i]);
}


void Geometry::setPeriodicity(const double periods[])
{
    periods_[0]=periods[0];
    periods_[1]=periods[1];
    periods_[2]=periods[2];
    inv_periods_[0] = 1.0/periods[0];
    inv_periods_[1] = 1.0/periods[1];
    inv_periods_[2] = 1.0/periods[2];
}



/** Minimum image representation.
    @return Minimum image representation of @a r. The value returned is @f$ \frac{-period}{2} \le r \le \frac{period}{2} @f$.
    The periods should be set with setPeriodicity();
    @param[in] r      Three-dimension vector.
    @see setPeriodicity().*/
void Geometry::applyPeriodicity1(double r[])
{
    applyPeriodicity0(r, periods_);
}

/** Minimum image representation.
    @return Minimum image representation of @a r. The value returned is @f$ \frac{-period}{2} \le r \le \frac{period}{2} @f$.
    The periods should be set with setPeriodicity();
    @param[in] r Cyclic coordinate.
    @param[in] axis   Either 0, 1 or 2.
    @see setPeriodicity().*/
double Geometry::applyPeriodicity1(double r, int const axis)
{
    return applyPeriodicity0(r, periods_[axis]);
}


/** Distance.
    @param[in] "x, y" Two three dimension vector.
    @param[out] z[]   
    @param[out] z1     @f$| \mathbf x -\mathbf y | @f$ with pbc applied.
    @see setPeriodicity().*/
void Geometry::distance(const double x[], const double y[], double z[], double & z1)
{
    double z2;
    distance(x, y, z, z1, z2);
}

/** Distance vector, norm.
    @see distance(const double x[], const double y[], double z[], double & z1, double & z2) and setPeriodicity().*/
void Geometry::distance(const double x[], const double y[], double z[])
{
    double z1,z2;
    distance(x, y, z, z1,z2);
}
/// Distance. @see distance
void Geometry::distance(const double x[], const double y[], double & z1)
{
    double z[3],z2;
    distance(x, y, z, z1,z2);
}

/** Distance vector, norm and norm square.
    @param[in] "x, y" Two three dimension vector.
    @param[out] z     Return x-y, |x-y| and |x-y|^2.            
    @see setPeriodicity().
*/

void Geometry::distance(const double x[], const double y[], Vector3 & z)
{
    distance(x, y, z.v, z._1, z._2);
}

/** Undo the separation of two atoms created by the periodic boundaries.
Example: two atoms H and O are at respectively x = -9.5 and x = 9.5 and the period along @em x is 20. The real distance between the two atoms is 1.
We use unBreak() on the coordinates of these two atoms:
@code 
unBreak(rh, ro);
@endcode
The coordinates @em rh of H is now 10.5 thus making O and H appears at a distance of 1 of each other.\n
Apply periodic boundaries condition to vector @f$ r - r_{ef} @f$.
@param[in]    r    Returns @f$ \textrm{applyPeriodicity}(r - r_{ef}) +r_{ef} @f$
@param[in] ref
@param[in] period
*/
double Geometry::unBreak0(double const r, double const ref, double const period)
{
    return applyPeriodicity0(r - ref, period)+ref;
}
/// Undo the separation of two atoms created by the periodic boundaries. @see unBreak0()
void Geometry::unBreak0(double r[], double const ref[], double const periods[])
{
    for (int a=0; a < 3; ++a)
        r[a]=unBreak0(r[a], ref[a], periods[a]);
}
/// Undo the separation of two atoms created by the periodic boundaries. @see unBreak0()
void Geometry::unBreak1(double r[], double const ref[])
{
    unBreak0(r, ref, periods_);
}
/// Undo the separation of two atoms created by the periodic boundaries. @see unBreak0()
double Geometry::unBreak1(double const r, double const ref, int const axis)
{
    return unBreak0(r, ref, periods_[axis]);
}


/** Calculate centre of two points.
Calculate the centre of @a r1 and @a r2 and store the result in @a rc.
@note The centre is calculated for the image or @a r2 the closed to @a r1.
*/
void Geometry::calculateCentre(double const r1[], double const r2[], double rc[])
{
    for (int i=0; i < 3; ++i) {
        rc[i]=(r1[i]+unBreak1(r2[i], r1[i], i))*0.5;
    };
}

/// Calculate centre of three points. @see calculateCentre().
void Geometry::calculateCentre(double const r1[], double const r2[], double const r3[], double rc[])
{
    for (int i=0; i < 3; ++i) {
        rc[i]=(r1[i]+unBreak1(r2[i], r1[i], i)+unBreak1(r3[i], r1[i], i))/3.0;
    };
}

/** Calculate barycentre.
The inverse of this functions the forces is spreadWeightedForce()
@pre Sum of weigths <em> w1, w2, w3 </em> must be 1.
*/
void Geometry::calculateWeightedCentre(double const w1, double const w2, double const w3, 
                                            double const r1[], double const r2[], double const r3[], double rc[])
{

    for (int i=0; i < 3; ++i) {
        rc[i]=( w1*r1[i]+w2*unBreak1(r2[i], r1[i], i)+w3*unBreak1(r3[i], r1[i], i) ) / ( w1+w2+w3);
    };
}

/** Calculate barycentre of diatomic system
*/
void Geometry::calculateWeightedCentre(double const w1, double const w2, double const r1[], double const r2[], double rc[]) {
	for(int i=0;i<3;++i) {
		rc[i]=( w1*r1[i]+w2*unBreak1(r2[i],r1[i],i) ) / (w1+w2);
	};
}



/** Spread force on centre to other points.
Spread force @a fc to @a f1, @a f2. Each receives a half of the force.
*/
void Geometry::spreadForce(double f1[], double f2[], double const fc[])
{
    for (int i=0; i < 3; ++i) {
        double const f=fc[i]/2.0;
        f1[i]+=f;
        f2[i]+=f;
    };
}

/** Spread force on centre to other points.
Spread force @a fc to @a f1, @a f2, @a f3. Each receives 1/3 of the force.
*/
void Geometry::spreadForce(double f1[], double f2[], double f3[], double const fc[])
{
    for (int i=0; i < 3; ++i) {
        double const f=fc[i]/3;
        f1[i]+=f;
        f2[i]+=f;
        f3[i]+=f;
    };
}

/** Spread force on barycentre to atoms.
This is the inverse functions calculateWeightedCentre()
@pre Sum of weigths <em> w1, w2, w3 </em> must be 1.
@param[in] "w1, w2, w3" weights.
@param[in,out] "f1, f2, f3" vector to spread @a fc to.
@param[in,out] fc force to spread. Is set to 0.0 by the function.
*/
void Geometry::spreadWeightedForce(const double  w1,const double  w2,const double  w3, 
                                        double f1[], double f2[], double f3[], double fc[])
{
    const double totalweightinv = 1.0/(w1+w2+w3);
    for (int i=0; i < 3; ++i) {
        double const f=fc[i]*totalweightinv;
        f1[i]+=f*w1;
        f2[i]+=f*w2;
        f3[i]+=f*w3;
        fc[i] =0;
    };

}

/** Spread force on barycentre to atoms.
This is the inverse functions calculateWeightedCentre()
@pre Sum of weigths <em> w1, w2</em> must be 1.
@param[in] "w1, w2" weights.
@param[in,out] "f1, f2" vector to spread @a fn to.
@param[in,out] fn force to spread.
*/
void Geometry::spreadWeightedForce( double const w1, double const w2, double f1[], double f2[], double fn[]) {
      double const winv=1.0/(w1+w2);
      for (int i=0; i < 3; ++i) {
            f1[i]+=fn[i]*w1*winv;
            f2[i]+=fn[i]*w2*winv;
            fn[i]=0.0;
      };
}

