/** @file
H2O-CO potential 
@author L.J. Karssemeijer
@date 2011
Radboud University of Nijmegen
*/

#include <iostream>
#include "h2oco.hpp"
#include "unit_system.hpp"
#include <iomanip>

using namespace h2ococo2forcefields;
using namespace h2ococo2forcefields::unit_system;



H2oCo::H2oCo() : PotentialBase() 
{ }

char const * H2oCo::getName()
{
      return "H2oCo";
}


void H2oCo::computeHH_O_C_O(const int nH2o , const int nCo, Water *const *const h2oList, Co *const *const coList, double & energy, const double * const diagbox) {


	setPeriodicity(diagbox);
	
	double qco_c, qco_o, r[3],r1,r2;
    Co    * co;
    Water * w;
    
    //double e=energy;
	//Loop over CO molecules first to minimize calculation of charges on CO (which is expensive due to the exponential form).
	for(int iCO=nCo-1; iCO >=0; --iCO) {
        co = coList[iCO];
        calcCO_charge(*co, qco_c, qco_o);

		for(int iH2O=nH2o-1; iH2O >=0; --iH2O) {
            w = h2oList[iH2O];
            if(co->fixed_ & w->fixed_) continue;

            distance(co->rx_, w->rx_, r,r1,r2);
			if(r1 <= cutoff_-switchingWidth_) { 
				 coulombFull(*co, *w, energy, qco_c, qco_o);
		         buckinghamFull(*co, *w, energy);

			}
		    //inside switching window

			else if(r1 < cutoff_) {
				double fw[4][3]={0.0}, fco[3][3]={0.0};
				double U=0.0;
					
				Co 	co_tmp(*co,fco[0],fco[1],fco[2]);
				Water w_tmp(*w, fw[0], fw[1], fw[2], fw[3]);
					
				coulombFull(co_tmp, w_tmp, U, qco_c, qco_o);
				buckinghamFull(co_tmp,w_tmp, U);
					
				//perform the switching.
				switching(co_tmp, w_tmp, U,r,r1);
				co->incrementForces(co_tmp);
				w->incrementForces(w_tmp);
				energy+=U;
			}

		}		
	}
  //  e = energy -e ;
  //  std::cout.precision(6);
  //  std::cout<<std::fixed<<"H2O-CO\n\tTotal: "<<e<<std::endl;    	
	
}


void H2oCo::computeInterMolecularH2oCo(Water &w, Co &co, double &energy,double *force,const double* const diagbox) {
    w.resetForces();
    co.resetForces();
    energy =0.0;
    double r[3],r1,r2,qco_c,qco_o;
    calcCO_charge(co,qco_c,qco_o);
    distance(w.rx_,co.rx_,r,r1,r2);
    
    if(r1>=cutoff_) return;
    buckinghamFull(co,w,energy);

    coulombFull(co,w,energy,qco_c,qco_o);

    if(r1 > ( cutoff_ - switchingWidth_ ) ) switching(co,w,energy,r,r1);

    for(int i=2;i>=0;--i) force[i]+=w.fh1_[i]+w.fh2_[i]+w.fo_[i] + w.fm_[i] + w.fx_[i];
    return;
}

void H2oCo::coulombFull(Co & co, Water & w, double &U, const double & qco_c, const double & qco_o) {

	const double qco_x=-qco_o-qco_c;
	
	double en;
	double r[3],r1,r2;

	double dqcdr[3],dqodr[3], dqxdr[3];
	
    //charge derivatives
    for(int i=0;i<3;i++) {
        dqcdr[i] = parameters::co_sigmac_ * co.d_[i] / co.d_1_;
        dqodr[i] = parameters::co_sigmao_ * co.d_[i] / co.d_1_;
        dqxdr[i] = ONE_OVER_4_PI_EPSILON0 * (parameters::co_sigmac_ * qco_c + parameters::co_sigmao_ * qco_o ) * co.d_[i] / co.d_1_;
    }
    
    //interaction between C and H1
    distance(co.rc_,w.rh1_,r,r1,r2);
    en = ONE_OVER_4_PI_EPSILON0*qco_c*parameters::tip4p_qh_ / r1 ;
    U+=en;
    for(int i=0;i<3;i++){
        w.fh1_[i] += - en * r[i] / r2;
        
        co.fc_[i] += en * r[i] / r2 + en * dqcdr[i];
        co.fo_[i] += - en * dqcdr[i]; 
    
    }
    //interaction between C and H2    
    distance(co.rc_,w.rh2_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0*qco_c*parameters::tip4p_qh_ / r1;
    U  += en;
    for(int i=0; i<3;i++) {
        w.fh2_[i] += - en * r[i] / r2;
        
        co.fc_[i] += en * r[i] / r2 + en * dqcdr[i]; 
        co.fo_[i] += - en * dqcdr[i];
    }
    //interaction between C and M
    distance(co.rc_,w.rm_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0*qco_c*parameters::tip4p_qm_ / r1;
    U  += en;
    for(int i=0; i<3;i++) {
        w.fm_[i]  += - en * r[i] / r2;
        
        co.fc_[i] += en * r[i] / r2 + en * dqcdr[i]; 
        co.fo_[i] += - en * dqcdr[i];
    }
    
    //interaction between O and H1
    distance(co.ro_,w.rh1_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0*qco_o*parameters::tip4p_qh_ / r1;
    U  += en;
    for(int i=0;i<3;i++) {
        w.fh1_[i] += - en * r[i] / r2; 
        
        co.fc_[i] += en * dqodr[i];     
        co.fo_[i] += en * r[i] / r2 - en * dqodr[i] ;
    }
    //interaction between O and H2
    distance(co.ro_,w.rh2_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0*qco_o*parameters::tip4p_qh_ / r1;
    U  += en;
    for(int i=0;i<3;i++) {
        w.fh2_[i] += - en * r[i] / r2; 
        
        co.fc_[i] += en * dqodr[i];     
        co.fo_[i] += en * r[i] / r2 - en * dqodr[i] ;
    }    
    //interaction between O and M
    distance(co.ro_,w.rm_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0*qco_o*parameters::tip4p_qm_ / r1;
    U  += en;
    for(int i=0;i<3;i++) {
        w.fm_[i]  += - en * r[i] / r2; 
        
        co.fc_[i] += en * dqodr[i];     
        co.fo_[i] += en * r[i] / r2 - en * dqodr[i] ;
    }
    
    //interaction between X and H1
    distance(co.rx_,w.rh1_,r,r1,r2);
    en = ONE_OVER_4_PI_EPSILON0 * qco_x * parameters::tip4p_qh_ / r1 ;
    U += en;
    for(int i=0;i<3;i++) {
        w.fh1_[i] += - en * r[i] / r2;
        
        co.fc_[i] += -dqxdr[i] * parameters::tip4p_qh_ / r1;
        co.fo_[i] +=  dqxdr[i] * parameters::tip4p_qh_ / r1;
        co.fx_[i] +=  en * r[i] / r2;
    }
    //interaction between X and H2
    distance(co.rx_,w.rh2_,r,r1,r2);
    en =  ONE_OVER_4_PI_EPSILON0 * qco_x * parameters::tip4p_qh_ / r1;
    U  += en;
    for(int i=0;i<3;i++) {
        w.fh2_[i] += - en * r[i] / r2;
        
        co.fc_[i] += -dqxdr[i] * parameters::tip4p_qh_ / r1;
        co.fo_[i] +=  dqxdr[i] * parameters::tip4p_qh_ / r1;
        co.fx_[i] += en * r[i] / r2;
    }
    //interaction between X and M
    distance(co.rx_,w.rm_,r,r1,r2);
    en = ONE_OVER_4_PI_EPSILON0 * qco_x * parameters::tip4p_qm_ / r1; 
    U += en; 
    for(int i=0;i<3;i++) {
        w.fm_[i]  += - en * r[i] / r2;
        
        co.fc_[i] += -dqxdr[i] * parameters::tip4p_qm_ / r1;
        co.fo_[i] +=  dqxdr[i] * parameters::tip4p_qm_ / r1;
        co.fx_[i] +=  en * r[i] / r2;
    }
}

void H2oCo::buckinghamFull(Co & co, Water & w, double &U) {

	buckingham(co.rc_, w.rh1_, co.fc_, w.fh1_, U, parameters::h2oco_csixch_, parameters::h2oco_cbuckch_, parameters::h2oco_betach_);
	buckingham(co.rc_, w.rh2_, co.fc_, w.fh2_, U, parameters::h2oco_csixch_, parameters::h2oco_cbuckch_, parameters::h2oco_betach_);
	buckingham(co.rc_, w.ro_ , co.fc_, w.fo_ , U, parameters::h2oco_csixco_, parameters::h2oco_cbuckco_, parameters::h2oco_betaco_);
	
	buckingham(co.ro_, w.rh1_, co.fo_, w.fh1_, U, parameters::h2oco_csixoh_, parameters::h2oco_cbuckoh_, parameters::h2oco_betaoh_);
	buckingham(co.ro_, w.rh2_, co.fo_, w.fh2_, U, parameters::h2oco_csixoh_, parameters::h2oco_cbuckoh_, parameters::h2oco_betaoh_);
	buckingham(co.ro_, w.ro_ , co.fo_, w.fo_ , U, parameters::h2oco_csixoo_, parameters::h2oco_cbuckoo_, parameters::h2oco_betaoo_);

}


void H2oCo::calcCO_charge(Co & co, double & qc, double & qo) {
	qc=parameters::co_qc_*exp( -parameters::co_sigmac_*(co.d_1_-parameters::co_re_) );
	qo=parameters::co_qo_*exp( -parameters::co_sigmao_*(co.d_1_-parameters::co_re_) );
}
