#include <iostream>
#include <iomanip>
#include "h2oco2.hpp"
#include "unit_system.hpp"
#include "helper_functions.hpp"

using namespace h2ococo2forcefields;
using namespace h2ococo2forcefields::unit_system;


H2oCo2::H2oCo2() :  qmqo(parameters::tip4p_qm_*parameters::co2_qo_),
                    qmqc(parameters::tip4p_qm_*parameters::co2_qc_),
                    qhqc(parameters::tip4p_qh_*parameters::co2_qc_),
                    qhqo(parameters::tip4p_qh_*parameters::co2_qo_)
 { };

char const * H2oCo2::getName() {
    return "H2oCo2";
}

void H2oCo2::computeH2O_CO2(const int nH2o, const int nCo2, Water *const *const h2oList, Co2 *const *const co2List, double &energy, const double* const diagbox) {
    setPeriodicity(diagbox);
    Water* w;
    Co2*   co2;
    double r[3],r1,r2;
    energy =0.0;
    double e= energy;

    for(int i=nH2o-1; i>=0;--i) {
        w = h2oList[i];
        for(int j=nCo2-1; j>=0;--j) {
            co2 = co2List[j];
            if(w->fixed_ & co2->fixed_) continue;        
    
            distance(w->rx_,co2->rx_,r,r1,r2);
            if(r1 <= cutoff_ - switchingWidth_) {
                fullInteraction(*w,*co2,energy);
            }
            else if(r1<= cutoff_) {
                double fco2_tmp[4][3]={0.0};                
                double fw_tmp[4][3]={0.0};                
                double U=0.0;
        
                Water w_tmp(*w, fw_tmp[0], fw_tmp[1], fw_tmp[2], fw_tmp[3]);
                Co2   co2_tmp(*co2, fco2_tmp[0], fco2_tmp[1], fco2_tmp[2], fco2_tmp[3]);
                
                fullInteraction(w_tmp,co2_tmp,U);
                switching(w_tmp,co2_tmp,U,r,r1);
            
                energy +=U;

                w->incrementForces(w_tmp);           
                co2->incrementForces(co2_tmp);
                                
            }        
        }
    }

    e = energy -e ;
    //std::cout.precision(6);
    //std::cout<<std::fixed<<"H2O-CO2\n\tTotal: "<<e<<std::endl;    

}

void H2oCo2::computeInterMolecularH2oCo2(Water &w, Co2 &co2, double &energy,double *force,const double* const diagbox) {
    double r[3],r1,r2;
    energy = 0.0;
    w.resetForces();
    co2.resetForces();    
    distance(w.rx_,co2.rx_,r,r1,r2);
    
    if(r1>=cutoff_) return;
    fullInteraction(w,co2,energy);
    if(r1>=(cutoff_-switchingWidth_)) switching(w,co2,energy,r,r1);
    for(int i=2;i>=0;--i) force[i]+=w.fh1_[i]+w.fh2_[i]+w.fo_[i] + w.fm_[i] + w.fx_[i];
    return;    

}

void H2oCo2::fullInteraction(Water &w, Co2 &co2,double &energy) {

    //electrostatics

    coulomb(w.rm_,co2.ro1_,w.fm_,co2.fo1_,energy,qmqo);
    coulomb(w.rm_,co2.ro2_,w.fm_,co2.fo2_,energy,qmqo);
    coulomb(w.rm_,co2.rc_,w.fm_,co2.fc_,energy,qmqc);

    coulomb(w.rh1_,co2.ro1_,w.fh1_,co2.fo1_,energy,qhqo);
    coulomb(w.rh1_,co2.ro2_,w.fh1_,co2.fo2_,energy,qhqo);
    coulomb(w.rh1_,co2.rc_,w.fh1_,co2.fc_,energy,qhqc);

    coulomb(w.rh2_,co2.ro1_,w.fh2_,co2.fo1_,energy,qhqo);
    coulomb(w.rh2_,co2.ro2_,w.fh2_,co2.fo2_,energy,qhqo);
    coulomb(w.rh2_,co2.rc_,w.fh2_,co2.fc_,energy,qhqc);

    //buckingham
    buckingham(w.ro_,co2.ro1_,w.fo_,co2.fo1_,energy,parameters::h2oco2_csixoo_,parameters::h2oco2_cbuckoo_,parameters::h2oco2_betaoo_);
    buckingham(w.ro_,co2.ro2_,w.fo_,co2.fo2_,energy,parameters::h2oco2_csixoo_,parameters::h2oco2_cbuckoo_,parameters::h2oco2_betaoo_);
    buckingham(w.ro_,co2.rc_, w.fo_,co2.fc_ ,energy,parameters::h2oco2_csixoc_,parameters::h2oco2_cbuckoc_,parameters::h2oco2_betaoc_);

    buckingham(w.rh1_,co2.ro1_,w.fh1_,co2.fo1_,energy,parameters::h2oco2_csixho_,parameters::h2oco2_cbuckho_,parameters::h2oco2_betaho_);
    buckingham(w.rh1_,co2.ro2_,w.fh1_,co2.fo2_,energy,parameters::h2oco2_csixho_,parameters::h2oco2_cbuckho_,parameters::h2oco2_betaho_);
    buckingham(w.rh1_,co2.rc_, w.fh1_,co2.fc_ ,energy,parameters::h2oco2_csixhc_,parameters::h2oco2_cbuckhc_,parameters::h2oco2_betahc_);

    buckingham(w.rh2_,co2.ro1_,w.fh2_,co2.fo1_,energy,parameters::h2oco2_csixho_,parameters::h2oco2_cbuckho_,parameters::h2oco2_betaho_);
    buckingham(w.rh2_,co2.ro2_,w.fh2_,co2.fo2_,energy,parameters::h2oco2_csixho_,parameters::h2oco2_cbuckho_,parameters::h2oco2_betaho_);
    buckingham(w.rh2_,co2.rc_ ,w.fh2_,co2.fc_ ,energy,parameters::h2oco2_csixhc_,parameters::h2oco2_cbuckhc_,parameters::h2oco2_betahc_);
}
