#include <iostream>
#include "h2ococo2pot.hpp"
#include <iomanip>

using namespace h2ococo2forcefields;

H2oCoCo2Pot::H2oCoCo2Pot() :	Potential(),
								Tip4p(),
								CoCo(),
								H2oCo(),
                                Co2Co2(),
                                H2oCo2(),
								initialized(false) {
}

H2oCoCo2Pot::~H2oCoCo2Pot() {
    std::cout<<"H2ococo2pot Destructor"<<std::endl;
    delete MolList;
}


void H2oCoCo2Pot::force(long N, const double *R, const int *atomicNrs, double *F, double *U, const double *box, const int * fixed) {

    diagbox[0] = box[0];
    diagbox[1] = box[4];
    diagbox[2] = box[8];
    if(!initialized) initialize(N,R,F,atomicNrs,box);
    
/*
    std::cout<<"n_h2o= "<<nH2O_molecules<<std::endl;
    std::cout<<"n_co = "<<nCO_molecules<<std::endl;
    std::cout<<"n_co2= "<<nCO2_molecules<<std::endl;
*/
        


    /******* RESET FORCES AND ENERGY ON ATOMS ************/

	*U=0.0;
	for(int i=0;i<3*N;++i) F[i]=0.0;

    /******* SET MOLECULES POSITIONS AND CALCULATE CENTRES OF MASS ************/

    MolList->initialize(R,F,fixed,diagbox);


    /******** CALCULATE INTERACTIONS ************/

    if(nH2O_molecules > 0) computeHH_O_(nH2O_molecules, MolList->h2oList, *U, diagbox);  

    if(nCO_molecules > 0)  computeC_O_(nCO_molecules, MolList->coList, *U, diagbox);

    double ebefore = *U;
    if(nCO2_molecules > 0) computeCO2_CO2(nCO2_molecules, MolList->co2List, *U, diagbox);
   // std::cout<<"CO2-CO2 interaction: "<< *U -ebefore <<std::endl;

    if(nH2O_molecules>0 && nCO_molecules>0) computeHH_O_C_O(nH2O_molecules,nCO_molecules, MolList->h2oList, MolList->coList, *U, diagbox);

    if(nH2O_molecules>0 && nCO2_molecules>0) computeH2O_CO2(nH2O_molecules,nCO2_molecules, MolList->h2oList, MolList->co2List, *U, diagbox); 


    /******** SPREAD FORCES TO ATOMS ************/
    Water   *w;
    Co      *co;  
    Co2     *co2;
    for(int i=nH2O_molecules-1; i>=0; --i) {
        w = MolList->h2oList[i];
        spreadForceRm(*w);
        spreadWeightedForce(parameters::weighto_, parameters::weighth_,  parameters::weighth_, w->fo_, w->fh1_, w->fh2_, w->fx_);
    }

    for(int i=nCO_molecules-1; i>=0; --i) {
        co = MolList->coList[i];
        spreadWeightedForce(parameters::weightc_, parameters::weighto_, co->fc_, co->fo_, co->fx_);	
    }

    for(int i=nCO2_molecules-1; i>=0; --i) {
        co2 = MolList->co2List[i];
        spreadWeightedForce(parameters::weighto_,parameters::weighto_, parameters::weightc_, co2->fo1_, co2->fo2_, co2->fc_, co2->fx_);
    }


    /******** CLEANUP MOLECULE LIST ************/
    MolList->cleanUp();
    


}


void H2oCoCo2Pot::intraMolecularForce(long N, const double *R, const int *atomicNrs, double *F, double *U, const double *box, const int * fixed) {
    diagbox[0] = box[0];
    diagbox[1] = box[4];
    diagbox[2] = box[8];
    MolList->initialize(R,F,fixed,diagbox);
    setPeriodicity(diagbox);
    initializeMolecularGeometries();    

    for(int i=3*N-1;i>=0;--i) F[i]=0.0;

    computeH2OIntra(nH2O_molecules,MolList->h2oList,*U,diagbox);    
    computeCO2Intra(nCO2_molecules,MolList->co2List,*U,diagbox);
    computeCOIntra(nCO_molecules,MolList->coList,*U,diagbox);

    MolList->cleanUp();    
    return;
}



void H2oCoCo2Pot::initialize(long N, const double *R, double *F, const int *atomicNrs,const double *box) {
    std::cout<<"initializing H2OCOCO2 potential class"<<std::endl;
    calc_nMolecules(N,atomicNrs);

    MolList = new Molecules(nH2O_molecules,nCO_molecules,nCO2_molecules,diagbox);

    initialized = true;
}

void H2oCoCo2Pot::calc_nMolecules(long N, const int *atomicNrs) {

    nH2O_molecules =0;
    nCO_molecules  =0;
    nCO2_molecules =0;

    //check if there are any water molecules.
    if(atomicNrs[0]==1) {
        //yes there are water molecules, we will count them now.
        nH2O_molecules=0;
        while(atomicNrs[2*nH2O_molecules]==1) nH2O_molecules+=1;
    }

    //if there are no more molecules we can stop here.
    if(N==3*nH2O_molecules) {
        return;
    }

    //now follows a list of carbon atoms followed by a number of oxygen atoms.
    unsigned int nC=0;
    unsigned int nO=0;
    unsigned int i = 3*nH2O_molecules;
    while(atomicNrs[i]==6) {
        i++;
        nC++;
    }
    while(atomicNrs[i]==8 && i < N) {
        i++;
        nO++;
    }
    //were these CO or CO2 molecules?
    if(nO == nC) {
        nCO_molecules = nC;
    }
    else if( nO == 2*nC) {
        nCO2_molecules = nC;
        return;
    }

    nCO2_molecules = ( N - 3*nH2O_molecules - 2*nCO_molecules) / 3;

    return;    
}

void H2oCoCo2Pot::listMolecules(const long nAtoms, int & nMolecules, int * nAtomsPerMolecule, int * atomIds) {

    int istart;


    if(initialized) {
        int iAtom=0;
        int iMol =0;
        nMolecules = nH2O_molecules+nCO_molecules+nCO2_molecules;

        for(int i=0;i<nH2O_molecules;i++) {
            nAtomsPerMolecule[iMol]=3;
            atomIds[iAtom] = 2*i; iAtom++;
            atomIds[iAtom] = 2*i+1; iAtom++;
            atomIds[iAtom] = 2*nH2O_molecules+i; iAtom++;
            iMol++;
        }
        for(int i=0;i<nCO_molecules;i++) {
            nAtomsPerMolecule[iMol]=2;
            atomIds[iAtom] = 3*nH2O_molecules + i; iAtom++;
            atomIds[iAtom] = 3*nH2O_molecules + nCO_molecules +i; iAtom++;
            iMol++;
        }
        for(int i=0;i<nCO2_molecules;i++) {
            nAtomsPerMolecule[iMol]=3;
            atomIds[iAtom] = 3*nH2O_molecules + 2*nCO_molecules + i; iAtom++;
            atomIds[iAtom] = 3*nH2O_molecules + 2*nCO_molecules + nCO2_molecules + 2*i; iAtom++;
            atomIds[iAtom] = 3*nH2O_molecules + 2*nCO_molecules + nCO2_molecules + 2*i +1 ; iAtom++;            
            iMol++;
        }
    }
    else {
        nMolecules = nAtoms;
        for(int i=nAtoms-1;i>=0;--i) {
            nAtomsPerMolecule[i]=1;
            atomIds[i]=i;
        }      
    }
    return;
}                
       
void H2oCoCo2Pot::initializeMolecularGeometries() {
    Water *w;
    Co *co;
    Co2 *co2;
    for(int i=nH2O_molecules-1;i>=0;--i) {
        w = MolList->h2oList[i];
        Tip4p::calculateInternalGeometry(*w);
    }
    for(int i=nCO_molecules-1;i>=0;--i) {
        co = MolList->coList[i];
        CoCo::calculateInternalGeometry(*co);
    }
    for(int i=nCO2_molecules-1;i>=0;--i) {
        co2 = MolList->co2List[i];
        Co2Co2::calculateInternalGeometry(*co2);
    }    
    return;
}

void H2oCoCo2Pot::dipoleMoment(double *D,const double *R, double *F, const double *box, const int * fixed) {

    diagbox[0] = box[0];
    diagbox[1] = box[4];
    diagbox[2] = box[8];
    MolList->initialize(R,F,fixed,diagbox);
    setPeriodicity(diagbox);
    initializeMolecularGeometries();   
    
    Water   *w;
    Co      *co;
    Co2     *co2;
    double r[3];
    double r1;
    
    for(int i=2;i>=0;--i) D[i] = 0.0;

    for(int i=nH2O_molecules-1; i>=0;--i) {
        w = MolList->h2oList[i];
        
        distance(w->rx_,w->rm_,r,r1);
        for(int j=2;j>=0;--j) D[j]+= parameters::tip4p_qm_ *(w->rx_[j] - r[j]);
        
        distance(w->rx_,w->rh1_,r,r1);
        for(int j=2;j>=0;--j)  D[j]+= parameters::tip4p_qh_ *(w->rx_[j] - r[j]);
        
        distance(w->rx_,w->rh2_,r,r1);
        for(int j=2;j>=0;--j) D[j]+= parameters::tip4p_qh_ *(w->rx_[j] - r[j]);                        
    }
    double qc,qo;
    for(int i=nCO_molecules-1;i>=0;--i) {
        co = MolList->coList[i];
        CoCo::calcCO_charge(*co,qc,qo);	        
        
        distance(co->rx_,co->rc_,r,r1);
        for(int j=2;j>=0;--j) D[j]+= qc * (co->rx_[j] - r[j]);
        
        distance(co->rx_,co->ro_,r,r1);
        for(int j=2;j>=0;--j)    D[j] += qo * (co->rx_[j] - r[j]);
    }
    for(int i=nCO2_molecules-1;i>=0;--i) {
        co2 = MolList->co2List[i];
        
        distance(co2->rx_,co2->rc_,r,r1);
        for(int j=2;j>=0;--j) D[j] += parameters::co2_qc_ * (co2->rx_[j] - r[j]);

        distance(co2->rx_,co2->ro1_,r,r1);
        for(int j=2;j>=0;--j) D[j] += parameters::co2_qo_ * (co2->rx_[j] - r[j]);

        distance(co2->rx_,co2->ro2_,r,r1);
        for(int j=2;j>=0;--j) D[j] += parameters::co2_qo_ * (co2->rx_[j] - r[j]);                
    }
    MolList->cleanUp();
    return;
}




void H2oCoCo2Pot::molecularVirial(long N, const double *R, const int *atomicNrs, double *F, double *U, const double *box, const int * fixed, double *virial) {



    diagbox[0] = box[0];
    diagbox[1] = box[4];
    diagbox[2] = box[8];
    MolList->initialize(R,F,fixed,diagbox);
    setPeriodicity(diagbox);
    initializeMolecularGeometries();    
    //calculates molecular virial 
    //All H2o-pairs
    double r[3],r1,rt[3];
    double f[3];
    double energy=0.0;
    
    for(int k=8;k>=0;--k) virial[k]=0.0;

    Water   *w1,*w2;
    Co      *co1,*co2;  
    Co2     *co21,*co22;
    //h2o-h2o
    for(int i=nH2O_molecules-1; i>=0;--i) {
        w1 = MolList->h2oList[i];
        for(int j=i-1;j>=0;--j) {
            for(int k=2;k>=0; --k) f[k]=0.0;
            w2 = MolList->h2oList[j];
            double fmag = 0.0;
            distance(w1->rx_,w2->rx_,r,r1);
            computeInterMolecularH2oH2o(*w1,*w2, energy,f,diagbox);

            for(int k=2;k>=0;--k)  {
                for(int l=k;l<3;l++)   { 
                    int index = 3*k+l;
                    virial[index]+= r[k]*f[l];
                }
            }
    }
    }
    
    //std::cout<<virial<<std::endl;
    //co2-co2
    for(int i=nCO2_molecules-1;i>=0;--i) {
        co21 = MolList->co2List[i];
        for(int j=i-1;j>=0;--j) {
            
            for(int k=2;k>=0; --k) f[k]=0.0;        
            
            co22 = MolList->co2List[j];
            distance(co21->rx_,co22->rx_,r);
            
            computeInterMolecularCo2Co2(*co21,*co22,energy,f,diagbox);

            for(int k=2;k>=0;--k)  {
                for(int l=k;l<3;l++)   { 
                    int index = 3*k+l;
                    virial[index]+= r[k]*f[l];
                }
            }

        }
    }
    //co-co
    for(int i=nCO_molecules-1;i>=0;--i) {
        co1=MolList->coList[i];
        for(int j=i-1;j>=0;--j){
            
            for(int k=2;k>=0;--k) f[k]=0.0;
            
            co2=MolList->coList[j];
            distance(co1->rx_,co2->rx_,r);
            
            
            computeInterMolecularCoCo(*co1,*co2,energy,f,diagbox);
            
            for(int k=2;k>=0;--k)  {
                for(int l=k;l<3;l++)   { 
                    int index = 3*k+l;
                    virial[index]+= r[k]*f[l];
                }
            }
            
        }
    }

    
    //h2o-co and h2o-co2
    //by this point, geometries of all molecules have been initialized.
    for(int i=nH2O_molecules-1; i>=0;--i) {
        w1 = MolList->h2oList[i];    
        //h2o-co
        for(int j=nCO_molecules-1;j>=0;--j) {
            co1 = MolList->coList[j];    
            for(int k=2;k>=0; --k) f[k]=0.0;
            distance(w1->rx_,co1->rx_,r,r1);
            computeInterMolecularH2oCo(*w1,*co1,energy,f,diagbox);
            
            for(int k=2;k>=0;--k)  {
                for(int l=k;l<3;l++)   { 
                    int index = 3*k+l;
                    virial[index]+= r[k]*f[l];
                }
            }
        }  
        //h2o-co2
        for(int j=nCO2_molecules-1;j>=0;--j) {
            co21 = MolList->co2List[j];   
            for(int k=2;k>=0; --k) f[k]=0.0;  
            
            distance(w1->rx_,co21->rx_,r);
            
            computeInterMolecularH2oCo2(*w1,*co21,energy,f,diagbox);
            
            for(int k=2;k>=0;--k)  {
                for(int l=k;l<3;l++)   { 
                    int index = 3*k+l;
                    //std::cout<<"adding "<<r[k]*f[l] <<std::endl;
                    virial[index]+= r[k]*f[l];
                }
            }
        }
    }

    for(int k=8;k>=0;--k) virial[k]*=0.5;
    virial[3]=virial[1];
    virial[6]=virial[2];
    virial[7]=virial[5];

    MolList->cleanUp();    
    return;
}
