#include "helper_functions.hpp"
#include <cmath>

namespace h2ococo2forcefields {
    namespace helper_functions {


/** Cross product.
@f$ \mathbf e=\mathbf v \times \mathbf w @f$
@param[in] v      3D vector.
@param[in]  w     3D vector.
@param[out] e     Result.
*/
double * crossProduct(const double v[], const double w[], double e[])
{
    e[0] = v[1] * w[2] - v[2] * w[1];
    e[1] = v[2] * w[0] - v[0] * w[2];
    e[2] = v[0] * w[1] - v[1] * w[0];
    return e;
}



/** Divide vector.
@param[in,out] v Three-dimension vector to divide.
@param[in] divisor Divide @a v by @a divisor.
*/
void divide(double v[], double const divisor)
{
    double factor = 1/divisor;
    for(int i=2;i>=0;--i) v[i]*=factor;
}

/** Dot product.
@f$ \mathbf v . \mathbf w @f$
@param[in] v      3D vector.
@param[in]  w     3D vector.
@return Dot product.
*/
double dotProduct(double const v[], double const w[])
{
    return v[0]*w[0]+v[1]*w[1]+v[2]*w[2];
}

/** Multiply vector.
@param[in,out] v Three-dimension vector to multiply.
@param[in] factor Multiply @a v by @a factor.
*/
void multiply(double v[], double const factor)
{
    v[0]*=factor;
    v[1]*=factor;
    v[2]*=factor;
}

/** Norm of 3D vector.
@f$ |\mathbf v| @f$
@param[in] v      3D vector.
@return Norm.
*/
double norm(double const v[])
{
    return sqrt(dotProduct(v, v));
}

/** Normalise 3D vector
@f$ \frac{\mathbf v}{ |\mathbf v|} @f$
@param[in,out] v  3D vector to normalise.
*/
void normalise(double v[])
{
    double const n=norm(v);
    divide(v, n);
}

}
}
