#include "molecules.hpp"
#include <iostream>
using namespace h2ococo2forcefields;

Molecules::Molecules(const int n_h2o,const int n_co,const int n_co2, const double * const diagbox) : nH2o(n_h2o), nCo(n_co), nCo2(n_co2)
{
    geom.setPeriodicity(diagbox);

    Rx= new double [ 3*(nH2o+nCo+nCo2) ];
    Fx= new double [ 3*(nH2o+nCo+nCo2) ];

    h2oList = new Water *[nH2o];
    coList =  new Co    *[nCo];
    co2List = new Co2   *[nCo2];

}

Molecules::~Molecules() {
    if(!cleanedUp) cleanUp();

    delete[] h2oList;
    delete[] coList;
    delete[] co2List;

    delete[] Rx;
    delete[] Fx;
}

void Molecules::initialize(const double* const R, double* const F, const int* const fixed, const double * const diagbox) {
    //set possible new box size
    geom.setPeriodicity(diagbox);

    //reset forces on centres of mass
    for(int i = 3*(nH2o+nCo+nCo2)-1; i>=0;--i) {
        Fx[i]=0.0;
    }

    //create molecule objects.
    initializeH2o(R,F,fixed);
    initializeCo(R,F,fixed);
    initializeCo2(R,F,fixed);

    cleanedUp=false;
}


void Molecules::cleanUp() {
    if(cleanedUp) return;

    for(int i=nH2o-1; i>=0;--i)  delete h2oList[i];
    for(int i=nCo-1; i>=0; --i)  delete coList[i];
    for(int i=nCo2-1; i>=0; --i) delete co2List[i];
    cleanedUp=true;
}

void Molecules::initializeH2o(const double* const R, double* const F, const int* const fixed) {
    int iMol, io, ih1, ih2;
    bool isfixed;

    for(int i=nH2o-1; i>=0;--i) {
        ih1 = 2*i;
        ih2 = 2*i+1;
        io  = 2*nH2o+i;
 
        isfixed = static_cast<bool>( fixed[ih1] & fixed[ih2] & fixed[io] );

        iMol = 3*i;
        ih1  = ih1*3;
        ih2  = ih2*3;
        io   = io*3;

        geom.calculateWeightedCentre(parameters::weighto_, parameters::weighth_, parameters::weighth_, &R[io], &R[ih1], &R[ih2], &Rx[iMol] );
        h2oList[i] = new Water(&R[ih1], &R[ih2], &R[io],&Rx[iMol], &F[ih1], &F[ih2], &F[io], &Fx[iMol], isfixed);
    }
}

void Molecules::initializeCo(const double* const R, double* const F, const int* const fixed) {
    int iMol,iStart,ic, io;
    iStart = 3*nH2o;
    bool isfixed;

    for(int i=nCo-1; i>=0; --i) {

        ic = iStart + i;
        io = iStart + nCo + i;

        isfixed = static_cast<bool>( fixed[ic] & fixed[io] );

        iMol    = 3*(nH2o+i);
        ic      =ic*3;
        io      =io*3;


        geom.calculateWeightedCentre(parameters::weightc_, parameters::weighto_, &R[ic], &R[io], &Rx[iMol] );
        coList[i] = new Co( &R[ic] , &R[io] , &Rx[iMol], &F[ic] , &F[io], &Fx[iMol], isfixed );
    }
}

void Molecules::initializeCo2(const double* const R, double* const F, const int* const fixed) {
    int iMol,iStart,ic,io1,io2;
    iStart = 3*nH2o+2*nCo;
    bool isfixed;

    for(int i=nCo2-1; i>=0; --i) {
        ic = iStart + i;
        io1 = iStart + nCo2 + 2*i;
        io2 = iStart + nCo2 + 2*i +1; 

        isfixed = static_cast<bool>( fixed[ic] & fixed[io1] & fixed[io2] );        

        iMol    = 3*(nH2o+nCo+i);
        ic      *=3;
        io1     *=3;
        io2     *=3;

       
        geom.calculateWeightedCentre(parameters::weightc_, parameters::weighto_, parameters::weighto_, &R[ic], &R[io1], &R[io2], &Rx[iMol]);
        co2List[i] = new Co2( &R[ic] , &R[io1] , &R[io2], &Rx[iMol] , &F[ic] , &F[io1], &F[io2], &Fx[iMol], isfixed);
    }
}
