/** @file
Basic  tools to write potentials.
@author Jean-Claude C. Berthet
@date 2006-2008
University of Iceland

Modified by L.J. Karssemeijer, 2011-2015
Radboud University Nijmegen
*/
#include <cmath>
#include <iostream>
#include <cassert>
#include <cstdlib>
#include "potential_base.hpp"
#include "helper_functions.hpp"
#include "unit_system.hpp"

/** @class h2ocoforcefields::PotentialBase
@brief Functions and tools commonly used by potentials.
Contains functions for quadratic retrainst, Lennard-Jones, Coulomb, to manage the periodic boundaries and interface for cutoff.
The system of units used is (eV, Angstrom, fs, e).
Rules to combine different Lennard-Jones parameters were taken from:

Unlike Lennard-Jones Parameters for Vapor-Liquid Equilibria, Thorsten Schnabel, Jadran Vrabec , Hans Hasse,
Institut fur Technische Thermodynamik und Thermische Verfahrenstechnik,
Universitat Stuttgart, D-70550 Stuttgart, Germany,
http://www.itt.uni-stuttgart.de/~schnabel/CR.pdf.
*/
using namespace std;
using namespace h2ococo2forcefields;
using namespace h2ococo2forcefields::unit_system;
using namespace h2ococo2forcefields::helper_functions;


/** Non bond interaction cutoff.
When the distance between two molecules is over getCutoff(), van der Waals and Coulomb interactions between the two molecules are ignored.
@see getSwitchingWidth().*/
PotentialBase::PotentialBase() :
Geometry(),
cutoff_(parameters::cutoff_),
switchingWidth_(parameters::switchingWidth_)
{
std::cout<<"PotentialBase constructor, no cutoff"<<std::endl;
}

PotentialBase::PotentialBase(double cutoff, double switchingWidth) :
Geometry(),
cutoff_(cutoff),
switchingWidth_(switchingWidth)
{
    if (switchingWidth_ > cutoff_) {
        cerr << "Error: getSwitchingWidth() > getCutoff()" << endl;
        exit(EXIT_FAILURE);
    };
	std::cout<<"PotentialBase constructor, cutoff specified"<<std::endl;
}

double PotentialBase::getCutoff() const
{
    return cutoff_;
}

void PotentialBase::setCutoff(double cutoff)
{
    cutoff_=cutoff;
}

/** Width of the switching zone.
In the switching zone the potential is replaced by @f$ V=U\times S(x) @f$ where @em U is the real potential @em V the switching potential and @em S the switching function. Variable x=0 at the beginning of the switching zone and x=1 at the end the function S is @f$ S=2x^3-3x^2+1 @f$ which has the following properties: S(0)=1, S(1)=0, S'(0)=0 and S'(1)=0.
The switching when the distance between the two molecules is getCutoff()-getSwitchingWidth() and ends when the distance is getCutoff();
@see getCutoff().
@{ */
double PotentialBase::getSwitchingWidth() const
{
    return switchingWidth_;
}

void PotentialBase::setSwitchingWidth(double width)
{
    switchingWidth_=width;
}
// @}


/** Calculate the base of an isosceles triangle.
An isosceles triangle is a triangle with two sides of equal length. If you know the length of the two equal sides and the angle between the two, the function can calculate the length of the third side.
@param[in]  length      Length of the two equal sides of the triangle.
@param[in]  angle       Angle between the two equal side in radiant.
*/
double PotentialBase::isoscelesBase(double length, double angle)
{
    return 2.0*length*sin(angle/2.0);
}

/** Compute quadratic restraints between two atoms.
@f$ U=k(r-r_0)^2 @f$
@param[in]              R1[]    Positions of atom 1.
@param[in]              R2[]    Positions of atom 2.
@param[in,out]       F1[]    Incremented by force on atom 1.
@param[in,out]       F2[]    Incremented by force on atom 2.
@param[in,out]       u      Incremented by energy.
@param[in]              k      Potential curvature
@param[in]              req    Distance of the two atoms at equilibrium.
@warning                 @a F1, @a F2 and @a u are incremented.*/      
void PotentialBase::restrainLength(const double R1[], const double R2[], double F1[], double F2[], double & u, const double k, const double req)
{
    double f, fx;
    double r[3], r1, r2;
    distance(R2, R1, r, r1, r2);
    double d=r1-req;
    f = -2.0*k*d;
    u+=k*d*d;
    fx=f*r[0]/r1; F1[0]-=fx; F2[0]+=fx;		
    fx=f*r[1]/r1; F1[1]-=fx; F2[1]+=fx;
    fx=f*r[2]/r1; F1[2]-=fx; F2[2]+=fx;
}

/** Angular quadratic restraints.
@f$ U=0.5*k*(theta-theta_eq) @f$
@param[in]  "rc,r1,r2"  Positions of the atoms.
@param[in,out] "fc,f1,f2" Forces on the atoms.
@param[in]  "k" force constant of the potential.
@param[in,out] "energy" energy.
@param[in] "theta_eq" equilibrium angle between atoms 1,c,2
*/
void PotentialBase::angleHarmonic(double const rc[], double const r1[], double const r2[], double fc[], double f1[], double f2[], double & energy, const double k, const double theta_eq) {
    double d1[3],d2[3],d1_1,d2_1;
    double dtheta,sintheta,costheta;
    double c11,c12,c22,P;
    
    distance(r1,rc,d1,d1_1,c11);
    distance(r2,rc,d2,d2_1,c22);

    c12 = dotProduct(d1,d2);

    costheta = c12 / d1_1 / d2_1;
    dtheta    = ( acos(costheta) - theta_eq );
    energy += 0.5*k*dtheta*dtheta;

    P = k*dtheta / ( sqrt( 1.0-costheta*costheta) * d1_1 *d2_1 );

    for(int i=2;i>=0;--i) {
        d1_1 = P * ( d2[i] - d1[i]*c12/c11);
        d2_1 = P * ( d1[i] - d2[i]*c12/c22);

        f1[i] += d1_1;
        f2[i] += d2_1;
        fc[i] -= d1_1 + d2_1;
    }

}

/** Compute Coulomb interaction between two charges.
@f$ U = \frac{q_1q_2}{|\mathbf r_1- \mathbf r_2|} @f$
@param[in]              r1	Positions of atom 1.
@param[in]              r2	Positions of atom 2.
@param[in,out]       f1	Incremented by force on atom 1.
@param[in,out]       f2	Incremented by force on atom 2.
@param[in,out]       energy           Incremented by energy.
@param[in]              qq        Product of the charges.
@warning	@a f1, @a f2 and @a u are incremented.*/
void PotentialBase::coulomb(const double r1[], const double r2[], double f1[], double f2[], double & energy, double qq)
{
    double d1, d2, d[3], f;
    distance(r1, r2, d, d1, d2);
    f = ONE_OVER_4_PI_EPSILON0*qq/d1;
    energy +=f;
    f = f / d2;
    for (int i=2; i >= 0; --i) {
        f1[i]+=f*d[i];
        f2[i]-=f*d[i];
    };
}


/** Lennard-Jones 12-6 between two atoms.
@param[in]              "R1, R2"	Positions of atom 1 and 2.
@param[in,out]       "F1, F2"	Force on atom 1 and 2.
@param[in,out]       E                    Energy.
@param[in]  "epsilon, sigma"  Parameters of the Lennard-Jones potential.
@warning	@a F1, @a F2 and @a u are incremented.
@see @ref lennardJones(double const, double &, double &, double const, double const) "lennardJones(distance, ...)".*/
void PotentialBase::lennardJones(const double R1[], const double R2[], double F1[], double F2[], double & E, double const epsilon, double const sigma)
{
    //WARNING: F1 and F2 are incremented.
    double F, en;
    double R12[3];
    double L, L2;
    distance(R2, R1, R12, L, L2);

    double x;
    x = sigma*sigma/L2;
    x*= x*x; 
    en = 4*epsilon*(x-1)*x;
    F = 24*epsilon*(2*x-1)*x/L2;
    E+=en;
    for(int i = 2; i>=0;--i) {
        F1[i] -=F*R12[i];
        F2[i] +=F*R12[i];
    }
}



/** Calculates forces and energy due to Buckingham interaction potential
	 The functional form of the potential is the following
	 @f[ 
	 V_{buck}=-\frac{C_6}{|r_i-r_j|^6}+C_b \exp{-\beta |r_i-r_j| }
	 @f]
	 */
void PotentialBase::buckingham(const double r1[], const double r2[], double f1[], double f2[], double & energy, const double csix, const double cbuck, const double beta) {
    double x[3],x1,x2,x6,expTerm,force;
	distance(r1,r2,x,x1,x2);    //r1 - r2
	
	x6      = x2*x2*x2;
	expTerm = exp(-beta*x1);
	force   = -6.0*csix/x6/x1 + cbuck*beta*expTerm;
	
	energy += cbuck*expTerm - csix/x6;
	for(int i=0;i<3;i++) {
	    f1[i] += force*x[i]/x1;
	    f2[i] += -force*x[i]/x1;
	}
}

/** Calculates forces and energy due to Buckingham interaction potential with Tang-Toennies damping
	 The functional form of the potential is the following
	 @f[ 
	 V_{buck}=-C_6\frac{f_6(\beta*|r_i-r_j|)}{|r_i-r_j|^6}+C_b \exp{-\beta |r_i-r_j| }
	 @f]
	 */
void PotentialBase::buckingham_tt(const double r1[], const double r2[], double f1[], double f2[], double & energy, const double csix, const double cbuck, const double beta) {

	double x[3],x1,x2,x6,TT,expTerm,force;
	distance(r1,r2,x,x1,x2);

	x6=x2*x2*x2;
	expTerm=exp(-beta*x1);
	tangToennies(beta*x1, 6, TT);	
	
	force= -6.0*csix*TT/x6/x1 + csix*pow(beta,7)*expTerm/720.0 + cbuck*beta*expTerm;

	energy+=cbuck*expTerm-csix*TT/x6;

	for(int i=0; i<3;i++) {
		f1[i] +=  force*x[i]/x1;
		f2[i] += -force*x[i]/x1;
	}

}
/** Calculates the Tang-Toennies damping function
	 The functional form of this function is the following
	 @f[
	 F_{N}(x)=1-\exp{-x}\sum_{n=0}^N \frac{x^n}{n!}	
	 @f]
	 @param[in]  x    function argument
	 @param[out] TT	return value.
	 @param[x]	 N		damping parameter
	*/
void PotentialBase::tangToennies(const double x, const int N, double & TT) {
	assert(N>=1);
	double term=1.0;
	double sum = 1.0;
	for(int i=1;i<=N;i++) {
		term*=x/i;
		sum += term;
	}
	TT=1.0-exp(-x)*sum;
}

/** Calculates morse potential between two atoms. 
	 The functional form of the potential is the following:
	 @f[
	 V(R)=D\left(1-\exp{-\beta(R-R_e)}\right)^2
	 @f]
	 @param[in] 	 R1,R2	Positions of the two atoms.
	 @param[in,out] F1,F2 	Forces on the atoms.
	 @param[in,out] u			Potential energy.
	 @param[in]		beta		potential parameter.
	 @param[in]	   D			potential parameter.
	 @param[in]		req		Equilibrium distance.
	 */
void PotentialBase::morse(const double R1[], const double R2[], double F1[], double F2[], double & u, const double beta, double D, const double req)
{
	double r[3], r1;
	distance(R1, R2, r, r1);	
    morse(R1,R2,F1,F2,r,r1,u,beta,D,req);
}


//d = R1 - R2
void PotentialBase::morse(const double R1[], const double R2[], double F1[], double F2[], const double d[], const double d1, double & u, const double beta, double D, const double req) {
    double f,fx;
	f = 1-exp(-beta*(d1-req));
	u+=D*f*f;
	f = -2.0*f*beta*D/d1;
	fx=f*d[0]; F1[0]+=fx; F2[0]-=fx;		
	fx=f*d[1]; F1[1]+=fx; F2[1]-=fx;
	fx=f*d[2]; F1[2]+=fx; F2[2]-=fx;    

}


/** Smooth cutoff between to Co molecules
	Should be called when the distance between the centers of mass of the two molecules is withing the switching range. 
	All forces on the atoms are decreased by a factor S (0<=S<=1). 
	The center of mass gets an additional force due to the derivative of the switching function w.r.t. the center of mass coordinate.
	This force should be spread to the C and O atoms through SpreadWeightedForce().
	@param[in,out] "co1,co2" the co molecules
	@param[in,out] "energy" energy due to the to-be-cut-off interaction.
	@note this function is for molecule-based cutoff.
	*/
void PotentialBase::switching(Co & co_1, Co & co_2, double & energy, const double *z, const double &z1) {
    double S,dS;
	switchingfunc(z1,S,dS);
    dS = energy *dS/z1;
    for(int i=2; i>=0;--i) {
    	co_1.fc_[i]*=S; co_1.fo_[i]*=S; co_1.fx_[i]*=S; 
    	co_2.fc_[i]*=S; co_2.fo_[i]*=S; co_2.fx_[i]*=S;

    	co_1.fx_[i]+= -dS*z[i];
    	co_2.fx_[i]+= dS*z[i]; 
    }
	energy *= S;
}

void PotentialBase::switching(Co2 & co2_1, Co2 & co2_2, double & energy, const double *z, const double & z1) {
    double  S, dS;
    switchingfunc(z1,S,dS);
    dS = energy *dS / z1;
    //multiply all forces by S;
    for(int i=2;i>=0;--i) {
        co2_1.fc_[i]*=S; co2_1.fo1_[i]*=S; co2_1.fo2_[i]*=S; co2_1.fx_[i]*=S;
        co2_2.fc_[i]*=S; co2_2.fo1_[i]*=S; co2_2.fo2_[i]*=S; co2_2.fx_[i]*=S;    

        co2_1.fx_[i]+= -dS*z[i];
        co2_2.fx_[i]+=  dS*z[i];
    }
    energy *=S;
}

void PotentialBase::switching(Water & w1, Water & w2, double & energy ,const double * z,const double & z1 ) {
    double S,dS;
    switchingfunc(z1,S,dS);
    dS = energy*dS/z1;
    for(int i=2; i>=0;--i) {
        w1.fo_[i]*=S; w1.fh1_[i]*=S; w1.fh2_[i]*=S; w1.fx_[i]*=S; w1.fm_[i]*=S;
        w2.fo_[i]*=S; w2.fh1_[i]*=S; w2.fh2_[i]*=S; w2.fx_[i]*=S; w2.fm_[i]*=S;

        w1.fx_[i] += -dS*z[i];
        w2.fx_[i] +=  dS*z[i];
    }
    energy*=S;
}

/** Smooth cutoff for the interactions of a CO and H2O molecule.
	Should be called when the distance between the centers of mass is in the switching range (otherwise assert will fail).
	All forces are decreased by a factor S.
	The center of mass gets an additional force due to the derivative of the switching function w.r.t. the center of mass coordinate. 
	This force should be spread out among the real atoms through SpreadWeightedForce().
	@param[in,out]	"co" CO molecule.
	@param[in,out] "w" H2O molecule.
	@param[in,out] "energy" potential energy. 
	@note this is a molecule-based cutoff. 
	*/
void PotentialBase::switching(Co & co, Water & w, double & energy, const double *z, const double &z1) {
	double S, dS, fxAdd;
	switchingfunc(z1,S,dS);
    dS = energy*dS/z1;
    for(int i=2; i>=0;--i) {
		co.fc_[i]*=S; co.fo_[i]*=S; co.fx_[i]*=S;
		w.fh1_[i]*=S; w.fh2_[i]*=S; w.fo_[i] *=S; w.fx_[i]*=S; w.fm_[i] *=S;
		
		co.fx_[i]+= -dS*z[i];
		w.fx_[i] +=  dS*z[i];
	}
	energy *= S;
}

void PotentialBase::switching(Water & w, Co2 & co2, double & energy, const double *z, const double &z1) {
    double S,dS,fxAdd;
    switchingfunc(z1,S,dS);
    dS = energy *dS /z1;
    for(int i=2; i>=0;--i) {
        co2.fc_[i] *=S; co2.fo1_[i]*=S; co2.fo2_[i]*=S; co2.fx_[i]*=S;
		w.fh1_[i]*=S; w.fh2_[i]*=S; w.fo_[i] *=S; w.fx_[i]*=S; w.fm_[i] *=S;

        w.fx_[i]   += -dS*z[i];
        co2.fx_[i] += dS*z[i];
    }
    energy *=S;
}


/** Switching function.
	Function to use to cut-off interactions beyond a certain range. 
	@param[in] "x_"		distance between the objects (must be inside the swiching range 
	@param[out] "S"	value of the switching function.
	@param[out] "dS"	derivative of the switching function.
	@note 
	*/
void PotentialBase::switchingfunc(const double & x_, double & S, double & dS) {
	double x=(x_-cutoff_+switchingWidth_)/switchingWidth_;
	assert(x>0.0);
	assert(x<1.0);
	S=(2*x-3)*x*x+1;
	dS=6*x*(x-1)/switchingWidth_;
}

void PotentialBase::spreadForceRm(Water & w) {
    double d1norm[3], d2norm[3];
    double P1,P22, P21,X1, X2;
    int i,j;

    P21 = 1.0/w.d1_1_;
    P22 = 1.0/w.d2_1_;
    for(i=2; i>=0; --i) {
        d1norm[i] = w.d1_[i] * P21; 
        d2norm[i] = w.d2_[i] * P22;
    }
   
    P1 = parameters::tip4p_dom_rel_ * 0.5;
    P21 = P1 * ( w.d1_1_ + w.d2_1_ ) * P21;
    P22 = P1 * ( w.d1_1_ + w.d2_1_ ) * P22;


    for(i=2; i>=0; --i) {
        //delta function terms
        w.fh1_[i] += w.fm_[i] *P21;
        w.fh2_[i] += w.fm_[i] *P22;
        w.fo_[i]  += w.fm_[i] *( 1.0 - P21 - P22 );

        X1 = w.fm_[i] * ( P1 * ( d1norm[i] + d2norm[i] ) - P21 * d1norm[i] );
        X2 = w.fm_[i] * ( P1 *  ( d1norm[i] + d2norm[i] ) - P22 * d2norm[i] );
        for(j=2; j>=0; --j) { 
            //forces in direction j on h1,h2,o, due to force on M in direction i.
            w.fh1_[j] += X1*d1norm[j];
            w.fh2_[j] += X2*d2norm[j];
            w.fo_[j] -= X1*d1norm[j] + X2*d2norm[j]; 
        }
        w.fm_[i] = 0.0;
    }
}
	

