#ifndef H2OCOCO2FORCEFIELDS_POTENTIAL_BASE_HPP_NEW
#define H2OCOCO2FORCEFIELDS_POTENTIAL_BASE_HPP_NEW

#include <cassert>
#include <cmath>
#include "water.hpp"
#include "co.hpp"
#include "co2.hpp"
#include "geometry.hpp"
#include "potential_parameters.hpp"
/** @file
Basic  tools to write potentials.
@author Jean-Claude C. Berthet
@date 2006-2008
University of Iceland

Modified by L.J. Karssemeijer, 2011-2015
Radboud University Nijmegen

*/
namespace h2ococo2forcefields {
    class PotentialBase : public h2ococo2forcefields::Geometry {
public:
        PotentialBase();
        PotentialBase(double cutoff, double switchingWidth);
        virtual ~PotentialBase(){};
        double getCutoff() const;
        void setCutoff(double cutoff);
        double getSwitchingWidth() const;
        void setSwitchingWidth(double width);
protected:

        //----------------------------------------Bond Pair interactions--------------------------------------------------
        /// @name Bond Interactions
        /// @{
        static double isoscelesBase(double length, double angle);
        void restrainLength(const double R1[], const double R2[], double F1[], double F2[], double & u, const double k, const double r0);
        /// @}
        void angleHarmonic(double const rc[], double const r1[], double const r2[], double fc[], double f1[], double f2[], double & energy, const double k, const double theta_eq);


        //----------------------------------------coulomb--------------------------------------------------      
        /// @name Coulomb
        /// @{
        //static const double ONE_OVER_4_PI_EPSILON0;
        void coulomb(const double r1[], const double r2[], double f1[], double f2[], double & u, double const qq);
        /// @}
                    
        //-------------------------------------------Lennard-Jones---------------------------------------------------------
        /// @name Lennard-Jones
        /// @{
        static double epsilon(double const A, double const B);
        void lennardJones(const double R1[], const double R2[], double F1[], double F2[], double & E, double const epsilon, double const sigma);
        //----------------------------------------Morse--------------------------------------------------      
        /// @name Morse
        /// @{      
        /** Interaction within the molecule
			Morse potential
			@param[in] R1 Position of atom 1
			@param[in] R2 Position of atom 2
			@param[in, out] F1 Add the force on atom 1 to @a F1
			@param[in, out] F2 Add the force on atom 2 to @a F2
			@param[in, out] u Add the potential energy to @a u
			@param[in] beta coefficient for Morse potential @f D*(1-exp(-beta(R-req)))^2
			@param[in] D    coefficient for Morse potential @f D*(1-exp(-beta(R-req)))^2
			@param[in] req  equilibrium distance for bond
			*/  
        void morse(const double R1[], const double R2[], double F1[], double F2[], double & u, const double beta, double D, const double req);
        void morse(const double R1[], const double R2[], double F1[], double F2[], const double d[], const double d1, double & u, const double beta, double D, const double req);

        //-----------------------------------------Buckingham---------------------------------------------------------
        /// @name Buckingham
        /// @{
        void buckingham_tt(const double r1[], const double r2[], double f1[], double f2[], double & energy, const double csix, const double cbuck, const double beta);
        void buckingham(const double r1[], const double r2[], double f1[], double f2[], double & energy, const double csix, const double cbuck, const double beta);
        
        void tangToennies(const double x, const int N, double & TT);
        /// @}
        

        //-------------------------------------------Switching---------------------------------------------------------      

        void switching(Water & w1, Water & w2, double & energy, const double * z, const double & z1 );
        void switching(Co2 & co2_1, Co2 & co2_2, double & energy,const double *z, const double & z1);
		void switching(Co & co1,Co & co2, double & energy, const double *z, const double & z1);

        void switching(Co & co, Water & w, double & energy,const double *z, const double & z1);
        void switching(Water & w, Co2 & Co2, double & energy, const double *z, const double & z1);        

        void switchingfunc(const double & x_, double & S, double & dS);

        

        //-----------------------------------------Spread force functions---------------------------------------------------------
        void spreadForceRm(Water & w);



        double cutoff_;
        double switchingWidth_;
        };
    }

#endif
