/** @file
Modified by L.J. Karssemeijer, 2011-2015
Radboud University Nijmegen
*/
#include <cmath>
#include <iostream>
#include <cassert>
#include "tip4p.hpp"

// If unit_system.hpp is removed the system of unit will be  (Energy: eV, distance: Angstrom, time: fs, charge: e).
#include "unit_system.hpp"

#include <iomanip>

using namespace std;
using namespace h2ococo2forcefields;

/* defined(FORCEFIELDS_UNIT_SYSTEM_HPP) && \
( FORCEFIELDS_UNIT_SYSTEM_HPP != FORCEFIELDS_UNIT_SYSTEM_ELECTRONVOLT_ANGSTROM_FEMTOSECOND_ECHARGE) //*/
using namespace h2ococo2forcefields::unit_system;

Tip4p::Tip4p() :
PotentialBase()
{ }

Tip4p::Tip4p(double cutoff, double switchingWidth) :
PotentialBase(cutoff, switchingWidth)
{ }



void Tip4p::computeH2OIntra(const int nH2o, Water *const *const h2oList, double & energy, const double* const diagbox) {
    Water *w1;
    for(int i=nH2o-1;i>=0;--i) {
        w1 = h2oList[i];
        calculateInternalGeometry(*w1);        
        if(w1->fixed_) continue;        
        computeIntraMolecularH2o(*w1,energy);
    }
    return;
}   

void Tip4p::computeHH_O_(const int nH2o, Water *const *const h2oList, double &energy, const double* const diagbox) {
    setPeriodicity(diagbox);

    Water *w1, *w2;
    double r[3], r1, r2;            

    double e_intra = energy;
    //first calculate intramolecular forces
    for(int i=nH2o-1; i>=0; --i) {
        w1 = h2oList[i];
        calculateInternalGeometry(*w1);
        if(w1->fixed_) continue;
        computeIntraMolecularH2o(*w1,energy);
    }
    e_intra = energy - e_intra;
    double e_inter = energy;
    //all pair contributions
    for(int i=nH2o-1; i>=0; --i) {
        w1 = h2oList[i];
        for(int j=i-1; j>=0; --j) {
            w2 = h2oList[j];
            if(w1->fixed_ & w2->fixed_) continue;
            
            distance(w1->rx_,w2->rx_,r,r1,r2);
            
            if(r1 <= cutoff_ - switchingWidth_) {
                //full interactions
                coulombFull(*w1,*w2,energy);
                lennardJones(w1->ro_, w2->ro_, w1->fo_, w2->fo_, energy, parameters::tip4p_epsilon_, parameters::tip4p_sigma_);
            }
            else if(r1 < cutoff_) {
                //switched interactions
                double f_tmp1[4][3]={0.0};
                double f_tmp2[4][3]={0.0};
                double U=0.0;
                Water w_tmp1(*w1,f_tmp1[0],f_tmp1[1],f_tmp1[2],f_tmp1[3]);    
                Water w_tmp2(*w2,f_tmp2[0],f_tmp2[1],f_tmp2[2],f_tmp2[3]);

                coulombFull(w_tmp1, w_tmp2,U);
                lennardJones(w_tmp1.ro_,w_tmp2.ro_, w_tmp1.fo_, w_tmp2.fo_, U , parameters::tip4p_epsilon_, parameters::tip4p_sigma_);

                switching(w_tmp1, w_tmp2, U,r,r1);

                w1->incrementForces(w_tmp1);
                w2->incrementForces(w_tmp2);

                energy +=U;                
            }
        }
    }
   /* 
    e_inter = energy - e_inter;
    std::cout.precision(6);
    std::cout<<std::fixed<<"H2O-H2O\n\tTotal: "<<e_inter+e_intra<<"\n\tInter: "<<e_inter<<"\n\tIntra: "<<e_intra<<std::endl;
*/


}

void Tip4p::computeIntraMolecularH2o(Water &w1, double &energy) {
        angleHarmonic(w1.ro_, w1.rh1_, w1.rh2_, w1.fo_, w1.fh1_, w1.fh2_, energy, parameters::tip4p_ktheta_, parameters::tip4p_theta_eq_);
        morse(w1.rh1_, w1.ro_, w1.fh1_, w1.fo_,w1.d1_, w1.d1_1_, energy, parameters::tip4p_beta_, parameters::tip4p_Dr_, parameters::tip4p_roh_eq_);
        morse(w1.rh2_, w1.ro_, w1.fh2_, w1.fo_,w1.d2_, w1.d2_1_, energy, parameters::tip4p_beta_, parameters::tip4p_Dr_, parameters::tip4p_roh_eq_);
        
        return;
}

void Tip4p::computeInterMolecularH2oH2o(Water &w1, Water &w2, double &energy,double *force,const double* const diagbox) {
    setPeriodicity(diagbox);
    double r[3], r1, r2;            
    energy = 0.0;
    w1.resetForces();
    w2.resetForces();

    distance(w1.rx_,w2.rx_,r,r1,r2);

    if(r1 > cutoff_) return;
    
    coulombFull(w1,w2,energy);
    lennardJones(w1.ro_, w2.ro_, w1.fo_, w2.fo_, energy, parameters::tip4p_epsilon_, parameters::tip4p_sigma_);    
    if(r1 > ( cutoff_ - switchingWidth_ ) ) switching(w1, w2, energy,r,r1);

    for(int i=2;i>=0;--i) force[i]+=w1.fo_[i]+w1.fh1_[i]+w1.fh2_[i]+w1.fx_[i]+w1.fm_[i];
}

void Tip4p::calculateInternalGeometry(Water & w) {
    double rm[3],d1[3],d2[3],d1_1,d2_1;

    distance(w.rh1_,w.ro_, d1, d1_1);
    distance(w.rh2_,w.ro_, d2, d2_1);


    double C = parameters::tip4p_dom_rel_ * 0.5 * (d1_1+d2_1); 
    for(int i=2;i>=0;--i) {
        rm[i] = w.ro_[i] + C*(d1[i]/d1_1 + d2[i]/d2_1);
    }
    w.setRm(rm);
    w.setD(d1,d2, d1_1, d2_1);
    distance(w.ro_, rm, d1_1);
}


inline void Tip4p::coulombFull(Water & w1, Water & w2, double & U)
{
    const double qhqh       = parameters::tip4p_qh_ * parameters::tip4p_qh_;
    const double qhqm       = parameters::tip4p_qh_ * parameters::tip4p_qm_;    

    // Coulomb interactions between hydrogens
    coulomb(w1.rh1_, w2.rh1_, w1.fh1_, w2.fh1_, U, qhqh );
    coulomb(w1.rh1_, w2.rh2_, w1.fh1_, w2.fh2_, U, qhqh );
    coulomb(w1.rh2_, w2.rh1_, w1.fh2_, w2.fh1_, U, qhqh );
    coulomb(w1.rh2_, w2.rh2_, w1.fh2_, w2.fh2_, U, qhqh );
    // interactions between H and M.
    coulomb(w1.rm_,    w2.rh1_,     w1.fm_,  w2.fh1_,  U, qhqm);
    coulomb(w1.rm_,    w2.rh2_,     w1.fm_,  w2.fh2_,  U, qhqm);
    coulomb(w1.rh1_,   w2.rm_,      w1.fh1_, w2.fm_,   U, qhqm);
    coulomb(w1.rh2_,   w2.rm_,      w1.fh2_, w2.fm_,   U, qhqm);

    // interactions between M1, M2
    coulomb(w1.rm_, w2.rm_, w1.fm_, w2.fm_, U, parameters::tip4p_qm_ * parameters::tip4p_qm_);

}

